<?php

namespace Ems\Repository;

use Doctrine\ORM\EntityManager;
use Doctrine\DBAL\Connection;
use Ems\Services\ImportacaoAFDService;
use Util\{ DateUtil, Repositories };

class ImportacaoAFDRepository {

    /** @var EntityManager */
    protected $entityManager;
    
    /** @var Connection */
    protected $connection;

    public function __construct(EntityManager $entityManager) {
        $this->connection    = $entityManager->getConnection();
        $this->entityManager = $entityManager;
    }

    public function getPeriodoIntervaloImp($nrorg){

        $query = 'SELECT TO_CHAR(PA.DTINICIALAPURACAO, \'DD/MM/YYYY\') DTINICIALAPURACAO, 
                         TO_CHAR(PA.DTFINALAPURACAO, \'DD/MM/YYYY\') DTFINALAPURACAO, 
                         (NVL( SUBSTR(PO.HRTEMPODESCANSODIARIO, 1, 2) + ( SUBSTR(PO.HRTEMPODESCANSODIARIO, 4, 2) / 60 ), 11)) HRTEMPODESCANSODIARIO
                    FROM PTO_PERIODOAPURACAO PA
                    JOIN GPE_PARAMORG PO ON PO.NRORG = PA.NRORG
                   WHERE PA.NRORG = :P_NRORG
                     AND TRUNC(SYSDATE) BETWEEN TRUNC(PA.DTINICIALAPURACAO) AND TRUNC(PA.DTFINALAPURACAO)';
        
        $params = array(
            'P_NRORG' => $nrorg,
        );

        return $this->connection->fetchAll($query, $params);
    }

    public function retornaNrPisHistoricoAtualPorPessoa(int $nrorg, int $nrpessoa, string $dtmescompetenc) : array {

        $params = array(
            'NRPESSOA'       => $nrpessoa,
            'NRORG'          => $nrorg,
            'DTMESCOMPETENC' => $dtmescompetenc,
        );
         
         $sql = <<<SQL
                    SELECT S.NRPISPASEPPES FROM GPE_PESSOAH S
                    WHERE S.NRORG = :NRORG
                      AND S.NRPESSOA = :NRPESSOA
                      AND S.DTMESCOMPETENC = (SELECT DTMESCOMPETENC
                                                FROM (SELECT SH.DTMESCOMPETENC
                                                        FROM GPE_PESSOAH SH
                                                       WHERE SH.NRORG = S.NRORG
                                                         AND SH.NRPESSOA = S.NRPESSOA
                                                         AND SH.DTMESCOMPETENC <= :DTMESCOMPETENC
                                                    ORDER BY SH.DTMESCOMPETENC DESC)
                                                WHERE ROWNUM = 1)
SQL;
         
         $result = $this->connection->fetchAll($sql, $params);
         return $result;
    }

    public function getMarcacoesVinculoMesCompetencia(int $nrOrg, string $dtIniStr, string $dtFimStr) : array {

        $query = $this->entityManager->getConnection()->prepare(
                 "SELECT NRVINCULOM, 
                         TO_CHAR(DTHRMARCACAOPONTO, 'DD/MM/YYYY HH24:MI') DTHRMARCACAOPONTO, 
                         IDNSRREP
                    FROM GPE_MARCACAOPONTO
                   WHERE NRORG = :P_NRORG
                     AND DTHRMARCACAOPONTO BETWEEN TO_DATE(:P_INI, 'DD/MM/YYYY HH24:MI:SS') AND TO_DATE(:P_FIM, 'DD/MM/YYYY HH24:MI:SS')"
        );
        
        $query->execute([
                "P_NRORG" => $nrOrg,
                "P_INI"   => $dtIniStr,
                "P_FIM"   => $dtFimStr
            ]);
        
        $retorno = $query->fetchAll();
        return $retorno;
    }

    /**
     *  Rotina que verifica se os intervalos de horário não incluem marcações 'DA' ou 'FA' que remetem à dias anteriores
     * 
     *  @param      $nrorg
     *  @param      $nrhordiam
     * 
     *  @return     Array   array('error' => $errorMesage, 'response' => boolean)
     */
    public function horarioIncluiDiaAnterior($nrorg, $nrhordiam): bool {
        $intervalosHorario = $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(array('nrorg' => $nrorg, 'nrhordiam' => $nrhordiam));
        if (empty($intervalosHorario)) {
            return false;
        }
        
        foreach($intervalosHorario as $intervaloHorario) {
            if (in_array($intervaloHorario->getIddiaocorreentrada(), ImportacaoAFDService::IDDIAOCORRENCIA_ANTERIORES) || in_array($intervaloHorario->getIddiaocorresaida(), ImportacaoAFDService::IDDIAOCORRENCIA_ANTERIORES)) {
                return true;
            }
        }
        
        return false;
    }

    /**
     *  Rotina que verifica se os intervalos de horário não incluem marcações 'DS' ou 'FS' que remetem à dias seguintes
     * 
     *  @param      $nrorg
     *  @param      $nrhordiam
     * 
     *  @return     Array   array('error' => $errorMesage, 'response' => boolean)
     */
    public function horarioIncluiDiaSeguinte($nrorg, $nrhordiam): bool {
        $intervalosHorario = $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(array('nrorg' => $nrorg, 'nrhordiam' => $nrhordiam));
        if (empty($intervalosHorario)) {
            return false;
        }
        
        foreach($intervalosHorario as $intervaloHorario) {
            if (in_array($intervaloHorario->getIddiaocorreentrada(), ImportacaoAFDService::IDDIAOCORRENCIA_SEGUINTES) || in_array($intervaloHorario->getIddiaocorresaida(), ImportacaoAFDService::IDDIAOCORRENCIA_SEGUINTES)) {
                return true;
            }
        }
        
        return false;
    }

    /**
     *  Rotina que verifica se os intervalos de horário não incluem marcações 'DC' ou 'FC' que remetem à dias correntes
     * 
     *  @param      $nrorg
     *  @param      $nrhordiam
     * 
     *  @return     Array   array('error' => $errorMesage, 'response' => boolean)
     */
    public function horarioIncluiDiaCorrente($nrorg, $nrhordiam): bool {
        $intervalosHorario = $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(array('nrorg' => $nrorg, 'nrhordiam' => $nrhordiam));
        if (empty($intervalosHorario)) {
            return false;
        }
        
        foreach($intervalosHorario as $intervaloHorario) {
            if (in_array($intervaloHorario->getIddiaocorreentrada(), ImportacaoAFDService::IDDIAOCORRENCIA_CORRENTES) || in_array($intervaloHorario->getIddiaocorresaida(), ImportacaoAFDService::IDDIAOCORRENCIA_CORRENTES)) {
                return true;
            }
        }
        
        return false;
    }

    private function getIntervalosHorario(string $nrOrg, string $nrHordiam): ?Array {
        return $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(
            array('nrorg' => $nrOrg, 'nrhordiam' => $nrHordiam),
            array('nrseqintervhorario' => 'ASC')
        );
    }
    
    private function getUltimoIntervaloHorario(string $nrOrg, string $nrHordiam): ?GpeIntervaloshorario {
        $intervalosHorario = $this->getIntervalosHorario($nrOrg, $nrHordiam);
        return count($intervalosHorario) > 0 ? end($intervalosHorario) : null;
    }

    private function getPrimeiroIntervaloHorario(string $nrOrg, string $nrHordiam): ?GpeIntervaloshorario {
        $intervalosHorario = $this->getIntervalosHorario($nrOrg, $nrHordiam);
        return count($intervalosHorario) > 0 ? reset($intervalosHorario) : null;
    }
    
    public function getHoraUltimaMarcacaoHorario(string $nrOrg, string $nrHordiam): ?float {
        $ultimoIntervalo = $this->getUltimoIntervaloHorario($nrOrg, $nrHordiam);
        return !empty($ultimoIntervalo) ? DateUtil::getHoraFloatDeString($ultimoIntervalo->getHrsaida()) : null;
    }

    public function getHoraPrimeiraMarcacaoHorario(string $nrOrg, string $nrHordiam): ?float {
        $primeiroIntervalo = $this->getPrimeiroIntervaloHorario($nrOrg, $nrHordiam);
        return !empty($primeiroIntervalo) ? DateUtil::getHoraFloatDeString($primeiroIntervalo->getHrentrada()) : null;
    }

    public function getDadosHorarioVinculo(int $nrorg, int $nrvinculom, string $dtapuracaoponto, ?string $dtmescompetenc) {
        
        $query = $this->entityManager->getConnection()->prepare("
            SELECT HT.NRHORDIAM, HD.IDTIPOHORARIO, IH.NRINTERVALOHORARIO, HH.NMHORDIAH, AE.NRESCALATRABM, EH.QTHRESCALATRABH AS QTJORNADADIARIA
		  FROM (SELECT NRORG, MAX(NRTURNO) NRTURNO, NRESCALATRABM
					 FROM GPE_ALTEESCALA
					WHERE NRORG = :P_NRORG
					  AND NRVINCULOM = :P_NRVINCULOM
					  --AND TO_DATE(:P_DTAPURACAOPONTO,'DD/MM/YYYY') BETWEEN DTINIESCALA AND NVL(DTFIMESCALA, TO_DATE(:P_DTAPURACAOPONTO, 'DD/MM/YYYY'))
                      AND DTINIESCALA = (SELECT MAX(DTINIESCALA)
                                 FROM GPE_ALTEESCALA ALT
                                WHERE ALT.NRORG        = GPE_ALTEESCALA.NRORG
                                  AND ALT.NRVINCULOM   = GPE_ALTEESCALA.NRVINCULOM
                                  AND ALT.DTINIESCALA <= TO_DATE(:P_DTAPURACAOPONTO,'DD/MM/YYYY'))
					GROUP BY NRORG, NRESCALATRABM) AE,
				 (SELECT NRORG, NRTURNO, TO_DATE(:P_DTAPURACAOPONTO,'DD/MM/YYYY')-DTDATABASE AS NRDIASENTRE
					 FROM GPE_TURNO
					WHERE NRORG = :P_NRORG) TU,
				 (SELECT NRORG, NRTURNO, NRHORDIAM,
							ROW_NUMBER() OVER (PARTITION BY NRORG, NRTURNO ORDER BY NRSEQHORARIO) AS NRSEQHR
					 FROM GPE_HORARIOTURNO
					WHERE NRORG = :P_NRORG
					ORDER BY NRORG, NRTURNO, NRSEQHORARIO) HT,
				 (SELECT NRORG, NRTURNO, COUNT(*) AS QTHORARIOS
					 FROM GPE_HORARIOTURNO
					WHERE NRORG = :P_NRORG
					GROUP BY NRORG, NRTURNO) QH,
				 (SELECT NRORG, NRHORDIAM, MAX(DTMESCOMPETENC) AS DTMESCOMPETENC, NMHORDIAH
					 FROM GPE_HORDIAH
					WHERE NRORG = :P_NRORG
					  AND DTMESCOMPETENC <= TO_DATE(:P_DTMESCOMPETENCIA, 'DD/MM/YYYY')
					GROUP BY NRORG, NRHORDIAM, NMHORDIAH) HH,
				 (SELECT NRORG, NRHORDIAM, MIN(NRSEQINTERVHORARIO) NRSEQINTERVHORARIO
					 FROM GPE_INTERVALOSHORARIO
					WHERE NRORG = :P_NRORG
					GROUP BY NRORG, NRHORDIAM) IV, GPE_INTERVALOSHORARIO IH, GPE_HORDIAH HD, GPE_ESCALATRABH EH
		 WHERE AE.NRORG = TU.NRORG
			AND AE.NRTURNO = TU.NRTURNO
			AND AE.NRORG = HT.NRORG
			AND AE.NRTURNO = HT.NRTURNO
			AND HT.NRORG = HH.NRORG
			AND HT.NRHORDIAM = HH.NRHORDIAM
			AND HD.NRORG = HH.NRORG
			AND HD.NRHORDIAM = HH.NRHORDIAM
			AND HD.DTMESCOMPETENC = HH.DTMESCOMPETENC
			AND IV.NRORG(+) = HH.NRORG
			AND IV.NRHORDIAM(+) = HH.NRHORDIAM
			AND IH.NRORG(+) = IV.NRORG
			AND IH.NRHORDIAM(+) = IV.NRHORDIAM
			AND IH.NRSEQINTERVHORARIO(+) = IV.NRSEQINTERVHORARIO
			AND AE.NRORG = QH.NRORG
			AND AE.NRTURNO = QH.NRTURNO
			AND HT.NRSEQHR-1 = MOD(TU.NRDIASENTRE, QH.QTHORARIOS)
      AND EH.NRESCALATRABM = AE.NRESCALATRABM
      AND EH.NRORG = AE.NRORG
      AND EH.DTMESCOMPETENC = (SELECT MAX(DTMESCOMPETENC)
                                 FROM GPE_ESCALATRABH
                                WHERE NRORG = EH.NRORG
                                  AND NRESCALATRABM = AE.NRESCALATRABM
                                  AND DTMESCOMPETENC <= TO_DATE(:P_DTMESCOMPETENCIA, 'DD/MM/YYYY'))
        ");

        $query->execute(
            array(
                "P_NRORG"            => $nrorg,
                "P_NRVINCULOM"       => $nrvinculom,
                "P_DTMESCOMPETENCIA" => is_null($dtmescompetenc) ? DateUtil::getPrimeiroDiaDoMes(DateUtil::getDataAtual())->format(DateUtil::FORMATO_BRASILEIRO) : $dtmescompetenc,
                "P_DTAPURACAOPONTO"  => $dtapuracaoponto,
            )
        );
        
        $dadosHorario = $query->fetchAll();

        if (isset($dadosHorario[0])) {
            return $dadosHorario[0];
        } else {
            return false;
        }
    }

    public function retornaUltimaAlteEscalaPorDataInicio(int $nrorg, int $nrvinculom, string $dtiniescala) : array {
        try {
            $query = <<<SQL
SELECT *
  FROM GPE_ALTEESCALA
 WHERE NRORG = :P_NRORG
   AND NRVINCULOM = :P_VINCULOM
   AND DTINIESCALA = ( SELECT MAX(DTINIESCALA)
                         FROM GPE_ALTEESCALA B
                        WHERE B.NRORG = :P_NRORG
                          AND B.NRVINCULOM = :P_VINCULOM
                          AND B.DTINIESCALA <= :P_DTINIESCALA )
SQL;
            $params = ['P_NRORG'       => $nrorg,
                       'P_VINCULOM'    => $nrvinculom,
                       'P_DTINIESCALA' => $dtiniescala];
            
            return $this->connection->fetchAll($query, $params);
        } catch (\Exception $e) {
            throw $e;
        }
    }

}