<?php

namespace Ems\Services;

use Doctrine\ORM\EntityManager;

use Ems\Util\Utils;
use Ems\Repository\ImportacaoAFDRepository;

use HCM\Geral\Entity\GpevinculomEntity;
use HCM\Util\{HCMParameters, Util, NovoCodigo };

use Util\{ DateUtil, Environment, Repositories };

use Zeedhi\Framework\Session\Session;

class ImportacaoAFDService {
    /** @param EntityManager */
    protected $entityManager;

    /** @param Environment */
    protected $environment;

    /** @param ImportacaoAFDRepository */
    protected $importacaoAFDRepository;

    protected $nrOrg;
    protected $nrOrgPadrao;
    protected $hcmParameters;
    protected $gpeVinculomEntity;
    protected $gpeMarcacaoponto;
    protected $ptoProgOcorrencia;

    // Tipos de Marcação (Coletores)
    const NRCOLETOR_MARCACAO_IMPORTADA_REP          = 26;
    const IMPORTACAOREP_DEFAULT_NRCOLETOR_DEFAULT   = self::NRCOLETOR_MARCACAO_IMPORTADA_REP;
    
    // Tipo de dia de ocorrência das marcações
    const IDDIAOCORRENCIA_DIA_ANTERIOR          = 'DA';
    const IDDIAOCORRENCIA_DIA_CORRENTE          = 'DC';
    const IDDIAOCORRENCIA_DIA_SEGUINTE          = 'DS';
    const IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR    = 'FA';
    const IDDIAOCORRENCIA_FORCA_DIA_CORRENTE    = 'FC';
    const IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE    = 'FS';

    const IMPORTACAO_TEMPO_TOLERANCIA = '02:00';
    const IMPORTACAO_TEMPO_TOLERANCIA_ANTERIOR = self::IMPORTACAO_TEMPO_TOLERANCIA;
    const IMPORTACAO_TEMPO_TOLERANCIA_SEGUINTE = self::IMPORTACAO_TEMPO_TOLERANCIA;
    
    const IMPORTACAO_DEFAULT_DIA  = self::IDDIAOCORRENCIA_FORCA_DIA_CORRENTE;
    const IMPORTACAO_DIA_CORRENTE = self::IDDIAOCORRENCIA_FORCA_DIA_CORRENTE;
    const IMPORTACAO_DIA_SEGUINTE = self::IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE;
    const IMPORTACAO_DIA_ANTERIOR = self::IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR;

    // Agrupações de dias de ocorrencia
    const IDDIAOCORRENCIA_ANTERIORES    = [self::IDDIAOCORRENCIA_DIA_ANTERIOR, self::IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR];
    const IDDIAOCORRENCIA_CORRENTES     = [self::IDDIAOCORRENCIA_DIA_CORRENTE, self::IDDIAOCORRENCIA_FORCA_DIA_CORRENTE];
    const IDDIAOCORRENCIA_SEGUINTES     = [self::IDDIAOCORRENCIA_DIA_SEGUINTE, self::IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE];

    //Dados REP Control iD
    const INIT_URL_REP     = 'https://';
    const URL_LOGIN_REP    = '/login.fcgi';
    const URL_GET_AFD_REP  = '/get_afd.fcgi';
    const USER_ADM_REP     = 'admin';
    const PASSWORD_ADM_REP = 'admin';

    static $memoizedFunctions;

    public function __construct(EntityManager $entityManager, Environment $environment, ImportacaoAFDRepository $importacaoAFDRepository) {
        $session = new Session();
        $priDiaMesAtual = DateUtil::getPrimeiroDiaDoMes(DateUtil::getDataAtual(true))->format('d/m/Y');

        $this->entityManager           = $entityManager;
        $this->environment             = $environment;
        $this->importacaoAFDRepository = $importacaoAFDRepository;
        $this->nrOrg             = (int) $this->environment->getNrOrg();
        $this->nrOrgPadrao       = (int) $this->environment->getNrOrgPadrao();

        $this->hcmParameters     = new HCMParameters($session, null, null, null, null);
        $this->hcmParameters->set('CDOPERADOR', null);
        $this->hcmParameters->set('dtmescompetenc', $priDiaMesAtual);
        $this->hcmParameters->set('dtcompetencia', $priDiaMesAtual);
        $this->hcmParameters->set('nrorgtrab', $this->nrOrg);

        $this->gpeVinculomEntity = new GpevinculomEntity($entityManager, $this->hcmParameters);
        $this->gpeMarcacaoponto  = Repositories::GPE_MARCACAOPONTO;
        $this->ptoProgOcorrencia = Repositories::PTO_PROGOCORRENCIA;

        self::$memoizedFunctions = [
            'getDadosHorarioVinculo' => Util::memoized('getDadosHorarioVinculo', $this->importacaoAFDRepository),
        ];
    }

    public function importacaoAFD($dataAFD) {
        try {
            set_time_limit(0);
            $arquivo = $dtIniObj = $dtFimObj = $tempoDescanco = $pisArquivo = $gruposPis = $linhasMarcacoes = $dtIniStr = $dtFimStr = $marcacoes = null;
            $arrRegistros = $arrVinculos = $mensagem = [];
            $pisVinculoNaoEcontrado = '';
            $countImp = $contMarc = 0;

            list($arquivo, $dtIniStr, $dtFimStr, $dtIniObj, $dtFimObj, $tempoDescanco) = $this->preparaArquivo($dataAFD);
            list($pisArquivo, $gruposPis, $linhasMarcacoes)                            = $this->filtraArquivo($arquivo);

            $this->retornaVinculos($gruposPis, $pisArquivo, $arrVinculos, $countImp);
            $this->gravaArrRegistros($linhasMarcacoes, $arrVinculos, $arrRegistros, $pisVinculoNaoEcontrado, $dtIniObj, $dtFimObj);

            if($pisVinculoNaoEcontrado)
                $mensagem['erro'] = 'Vínculo não encontrado para o(s) pis: '.$pisVinculoNaoEcontrado.'.';
            
            $marcacoes = $this->importacaoAFDRepository->getMarcacoesVinculoMesCompetencia($this->nrOrg, $dtIniStr, $dtFimStr);

            $this->preparaRegistros($arrRegistros, $marcacoes, $contMarc, $tempoDescanco);

            $mensagem['sucesso'] = 'Importação de '.$contMarc.' marcações concluída para '.$countImp.' vínculos.';

            return $mensagem;

        } catch (\Exception $e) {
            throw new \Exception("Falha na importação AFD. Erro: ". $e->getMessage());
        }
    }

    private function preparaArquivo(string $dataAFD) : array {
        $peridApur =  $this->importacaoAFDRepository->getPeriodoIntervaloImp($this->nrOrg)[0];

        if (!$peridApur) {
            throw new \Exception("Periodo de apuração não encontrado.");
        } else {
            $dtIniStr      = $peridApur['DTINICIALAPURACAO'].' 00:00:00';   // Adiquiri as datas de escopo e ajusta a hora limite do dia
            $dtFimStr      = $peridApur['DTFINALAPURACAO']  .' 23:59:59';
            $dtIniObj      = DateUtil::getDataDeString($dtIniStr, DateUtil::FORMATO_BRASILEIRO_DATAHORA);  // Converte as datas de escopo para objeto
            $dtFimObj      = DateUtil::getDataDeString($dtFimStr, DateUtil::FORMATO_BRASILEIRO_DATAHORA);  // Converte as datas de escopo para objeto
            $tempoDescanco = $peridApur['HRTEMPODESCANSODIARIO'];
            $arrDataAFD    = explode(PHP_EOL, $dataAFD);
        }
        
        $arquivo = Util::filterREPFileImp($arrDataAFD, $dtIniObj, $dtFimObj);

        if($arquivo)
            return [$arquivo, $dtIniStr, $dtFimStr, $dtIniObj, $dtFimObj, $tempoDescanco];
        else
            throw new \Exception("Ocorreu um erro na preparação do arquivo.
                                 O período de apuração localizado com base na data atual não corresponde ao arquivo AFD que está sendo importado.");
    }

    private function filtraArquivo(array $arquivo) : array {
        $linhasMarcacoes = $pisArquivo = $gruposPis = null;
        
        /* Filtro do registros do tipo 3 */
        $linhasMarcacoes = array_filter(
            $arquivo, function ($linha) {
                $tpReg = substr($linha, 9, 1);
                return $tpReg == '3';
            }
        );

        $pisArquivo = array_map(
            function ($linha) {
                $pis = intval(substr($linha, 22, 12));
                return $pis;
            }, $linhasMarcacoes
        );

        $pisArquivo = array_unique($pisArquivo);
        $gruposPis  = array_chunk($pisArquivo, 500);
        
        return [$pisArquivo, $gruposPis, $linhasMarcacoes];
    }

    private function retornaVinculos(array $gruposPis, array $pisArquivo, array &$arrVinculos, int &$countImp) : void {
        $vinculosIndexado = [];

        foreach ($gruposPis as $grupoPis) {
            $novosVinculos = $this->gpeVinculomEntity->retornaVinculosAtivosMesCompetencia($this->nrOrg, $this->hcmParameters->getDtCompetencia(), null, null, null, false, null, true, false, null, null, null,  null, $pisArquivo);
            $arrVinculos = array_merge($arrVinculos, $novosVinculos);
        }

        foreach ($arrVinculos as $vinculo) {
            $pessoaVinculo = $this->importacaoAFDRepository->retornaNrPisHistoricoAtualPorPessoa($this->nrOrg, $vinculo->getNrpessoa(), $this->hcmParameters->getDtCompetencia())[0];

            if (empty($pessoaVinculo)) {
                continue;
            }
            
            $countImp++;
            $vinculosIndexado[$pessoaVinculo['NRPISPASEPPES']] = $vinculo;
        }

        $arrVinculos = $vinculosIndexado;
        unset($vinculosIndexado);
    }

    private function gravaArrRegistros (array $linhasMarcacoes, array $arrVinculos, array &$arrRegistros, string &$pisVinculoNaoEcontrado, \DateTime $dtIniObj, \DateTime $dtFimObj) : void {
        foreach ($linhasMarcacoes as $linha) {
            $nsr  = substr($linha, 0, 9); // Número Sequencial de Registro
            $data = substr($linha, 10,  8);
            $hora = substr($linha, 18,  4);
            $pis  = intval(substr($linha, 22, 12));
            $dtMarcacaoStr = substr($data, 0, 2) . '/' . substr($data, 2, 2) . '/' . substr($data, 4, 4)  //Monta a data de marcação completa (em string). Parte da data (string)
                           . ' ' . substr($hora, 0, 2) . ':' . substr($hora, 2, 2) . ':00';               //Parte da hora (string)

            $dtMarcacaoObj = DateUtil::getDataDeString($dtMarcacaoStr, DateUtil::FORMATO_BRASILEIRO_DATAHORA);

            $vinculo = $arrVinculos[$pis] ?? null;
            
            if(empty($vinculo)) {
                $pisVinculoNaoEcontrado .= ''.$pis.',';
                continue;
            }

            if(DateUtil::getIntervaloEmSegundos($dtIniObj, $dtMarcacaoObj) >= 0 && DateUtil::getIntervaloEmSegundos($dtMarcacaoObj, $dtFimObj) > 0) {
                $arrRegistros[] = [
                    'vinc'=> $vinculo->getNrvinculom(),
                    'mar' => $dtMarcacaoObj,
                    'nsr' => $nsr];
            }
        }
    }

    private function preparaRegistros(array $arrRegistros, array $marcacoes, int &$contMarc, string &$tempoDescanco) : void {
        
        foreach($arrRegistros as $valor){

            $marcAnterior     = $diaAnterior = $idDiaAnterior = $diaAtual = $dif = $nrMarcacaoPonto = $index = $horarioMarcacoes = null;
            $idDia            = self::IMPORTACAO_DEFAULT_DIA;
            $quebraIteracao   = false;
            $nrvinculom       = $valor['vinc'];
            $nsr              = $valor['nsr'];
            $diaAtual         = $valor['mar']->format('d/m/Y');

            //Procura marcações ja feitas ou importadas, se a importação verificar que ja existem com o mesmo nsr e o horario e a data, define a variavel de quebra de iteração como true.
            if (isset($marcacoes)) {
                foreach($marcacoes as $marcacao) {
                    if ($marcacao["NRVINCULOM"] == $nrvinculom) {
                        if ($marcacao["IDNSRREP"] == $nsr && $marcacao["DTHRMARCACAOPONTO"] == $valor['mar']->format('d/m/Y H:i')) {
                            $quebraIteracao = true;
                        }
                    }
                }
            }
            
            if ($quebraIteracao == true) {
                continue;
            }

            if(isset($marcAnterior)){
                // Condição para tornar a primeira marcação do dia como 'FA' caso a diferença dessa pra anterior seja inferior ao tempo de descanso (param. org.) e não seja no mesmo dia da anterior
                $dif = $valor['mar']->diff($marcAnterior);
                $dif = ($dif->s / 3600) + ($dif->i / 60) + $dif->h + ($dif->d * 24) + ($dif->m * 720) + ($dif->y * 8760);
                if($dif > $tempoDescanco) {
                    $idDia = self::IMPORTACAO_DIA_CORRENTE;
                } else {
                    if($diaAtual == $diaAnterior) {
                        $idDia = $idDiaAnterior;
                    } else {
                        $idDia = self::IMPORTACAO_DIA_ANTERIOR;
                    }
                }
            }

            $idDia = $this->getIdDiaMarcImportada($nrvinculom, $valor['mar']);
            
            $nrMarcacaoPonto = \HCM\Util\NovoCodigo::geraCodigo($this->nrOrg, 'GPE_MARCACAOPONTO', 1, 12, true);
            $marcAnterior    = $valor['mar'];
            $diaAnterior     = $marcAnterior->format('d/m/Y');
            $idDiaAnterior   = $idDia;
            
            $index = $idDia == 'FA' ? DateUtil::subtraiIntervalo($valor['mar'], 1)->format('d/m/Y') : $valor['mar']->format('d/m/Y');
            $horarioMarcacoes[$index][] = ['dthrmarcacaoponto' => $valor['mar'],
                                           'nrmarcacaoponto'   => $nrMarcacaoPonto,
                                           'iddiaocorrencia'   => $idDia];
            
            $this->insereGpeMarcacaoponto($nrvinculom, $marcAnterior, $nsr, $idDia, $nrMarcacaoPonto);
            $this->trataHorarioMarcacoes($horarioMarcacoes, $nrvinculom, $nrMarcacaoPonto, $valor);

            $contMarc++;
        }
    }

    /**
     * De acordo com os intervalos cadastrados para o horario previsto para o dia da batida, decide a sinalizacao da marcacao
     *
     * @return string (IMPORTACAO_DIA_CORRENTE, IMPORTACAO_DIA_SEGUINTE, IMPORTACAO_DIA_ANTERIOR)
     */
    private function getIdDiaMarcImportada(string $nrVinculo, \DateTime $dtHrMarc): string {
        /* Deteccao de virada para o dia anterior */
        
        $nrHordiamVinculoDiaAnterior = $this->getNrHordiamVinculoDiaAnterior($nrVinculo, $dtHrMarc);
        
        if($this->isMarcImportadaDiaAnterior($dtHrMarc, $nrHordiamVinculoDiaAnterior)) {
            return self::IMPORTACAO_DIA_ANTERIOR;
        }
        
        /* Deteccao de virada para o dia seguinte */
        
        $nrHordiamVinculoDiaSeguinte = $this->getNrHordiamVinculoDiaSeguinte($nrVinculo, $dtHrMarc);
        
        if($this->isMarcImportadaDiaSeguinte($dtHrMarc, $nrHordiamVinculoDiaSeguinte)) {
            return self::IMPORTACAO_DIA_SEGUINTE;
        }
        
        /* Dia convencional */
        return self::IMPORTACAO_DEFAULT_DIA;
    }

    /**
     * Busca o horario do dia imediatamente anterior ao da batida
     * 
     * @param string    $nrVinculo
     * @param \DateTime $dtHrMarc
     * 
     * @return ?string
     */
    private function getNrHordiamVinculoDiaAnterior(string $nrVinculo, \DateTime $dtHrMarc): ?string {
        $getDadosHorarioVinculoMemoized = self::$memoizedFunctions['getDadosHorarioVinculo'];

        $dadosHorarioVinculo = $getDadosHorarioVinculoMemoized(
            $this->hcmParameters->getNrorgTrabalho(),
            $nrVinculo,
            DateUtil::subtraiIntervalo($dtHrMarc, 1)->format(DateUtil::FORMATO_DATA),
            $this->hcmParameters->getDtCompetencia()
        );
        
        return $dadosHorarioVinculo['NRHORDIAM'] ?? null;
    }

    /**
     * Busca o horario do dia imediatamente seguinte ao da batida
     * 
     * @param string    $nrVinculo
     * @param \DateTime $dtHrMarc
     * 
     * @return ?string
     */
    private function getNrHordiamVinculoDiaSeguinte(string $nrVinculo, \DateTime $dtHrMarc): ?string {
        $getDadosHorarioVinculoMemoized = self::$memoizedFunctions['getDadosHorarioVinculo'];

        $dadosHorarioVinculo = $getDadosHorarioVinculoMemoized(
            $this->hcmParameters->getNrorgTrabalho(),
            $nrVinculo,
            DateUtil::adicionaIntervalo($dtHrMarc, 1)->format(DateUtil::FORMATO_DATA),
            $this->hcmParameters->getDtCompetencia()
        );
        
        return $dadosHorarioVinculo['NRHORDIAM'] ?? null;
    }

    /**
     * Verifica se a marcacao sera sinalizada como anterior considerando o horario do dia anterior
     * 
     * @param \DateTime $dtHrMarc
     * @param ?string   $nrHordiam
     * 
     * @return ?string
     */
    private function isMarcImportadaDiaAnterior(\DateTime $dtHrMarc, ?string $nrHordiam): bool {
        $naoExisteViradaNoHorario = !$this->importacaoAFDRepository->horarioIncluiDiaAnterior($this->hcmParameters->getNrorgTrabalho(), $nrHordiam);
            
        $horarioNaoInformado = empty($nrHordiam);
        
        if ($naoExisteViradaNoHorario || $horarioNaoInformado) {
            return false;
        }
        
        return $this->isMarcImportadaDentroToleranciaDiaAnterior($dtHrMarc, $nrHordiam);
    }

    /**
     * Verifica se a marcacao esta dentro do periodo previsto para ser sinalizada como anterior, considerando um tempo definido de tolerancia a partir da ultima batida do horario
     * 
     * @param \DateTime $dtHrMarc
     * @param string    $nrHordiam
     * 
     * @return bool
     */
    private function isMarcImportadaDentroToleranciaDiaAnterior(\DateTime $dtHrMarc, string $nrHordiam): bool {
        $horaUltimaMarcacao = $this->importacaoAFDRepository->getHoraUltimaMarcacaoHorario($this->hcmParameters->getNrorgTrabalho(), $nrHordiam);
        $ultimaMarcacaoDia = $this->getMarcacaoFromHrIntervalo($dtHrMarc, $horaUltimaMarcacao);
        
        $toleranciaDiaAnterior = DateUtil::adicionaIntervalo(
            $ultimaMarcacaoDia,
            DateUtil::getHoraFloatDeString(self::IMPORTACAO_TEMPO_TOLERANCIA_ANTERIOR),
            DateUtil::HORAS
        );
        
        return $dtHrMarc <= $toleranciaDiaAnterior;
    }

    /**
     * Soma a hora a data informada
     * 
     * @param \DateTime $dtHrMarc
     * @param ?float    $hora
     * 
     * @return \DateTime
     */
    private function getMarcacaoFromHrIntervalo(\DateTime $dtHrMarc, ?float $hora): \DateTime {
        $dataFinal = DateUtil::adicionaIntervalo(
            DateUtil::truncateData($dtHrMarc),
            floor($hora),
            DateUtil::HORAS
        );
        
        $dataFinal = DateUtil::adicionaIntervalo(
            $dataFinal,
            round(($hora - floor($hora)) * 60), // Decimal para minutos
            DateUtil::MINUTOS
        );
        
        return $dataFinal;
    }

    /**
     * Verifica se a marcacao sera sinalizada como seguinte considerando o horario do dia seguinte
     * 
     * @param \DateTime $dtHrMarc
     * @param ?string   $nrHordiam
     * 
     * @return ?string
     */
    private function isMarcImportadaDiaSeguinte(\DateTime $dtHrMarc, ?string $nrHordiam): bool {
        $naoExisteViradaNoHorario = !$this->importacaoAFDRepository->horarioIncluiDiaSeguinte($this->hcmParameters->getNrorgTrabalho(), $nrHordiam);
            
        $horarioNaoInformado = empty($nrHordiam);
        
        if ($naoExisteViradaNoHorario || $horarioNaoInformado) {
            return false;
        }
        
        return $this->isMarcImportadaDentroToleranciaDiaSeguinte($dtHrMarc, $nrHordiam);
    }

    /**
     * Verifica se a marcacao esta dentro do periodo previsto para ser sinalizada como seguinte, considerando um tempo definido de tolerancia a partir da primeira batida do horario
     * 
     * @param \DateTime $dtHrMarc
     * @param string    $nrHordiam
     * 
     * @return bool
     */
    private function isMarcImportadaDentroToleranciaDiaSeguinte(\DateTime $dtHrMarc, string $nrHordiam): bool {
        $horaPrimeiraMarcacao = $this->importacaoAFDRepository->getHoraPrimeiraMarcacaoHorario($this->hcmParameters->getNrorgTrabalho(), $nrHordiam);
        $primeiraMarcacaoDia = $this->getMarcacaoFromHrIntervalo($dtHrMarc, $horaPrimeiraMarcacao);
        
        $toleranciaDiaSeguinte = DateUtil::subtraiIntervalo(
            $primeiraMarcacaoDia,
            DateUtil::getHoraFloatDeString(self::IMPORTACAO_TEMPO_TOLERANCIA_SEGUINTE),
            DateUtil::HORAS
        );
        
        /* Caso o horario seja somente no dia seguinte ou no dia anterior */
        
        if (!$this->importacaoAFDRepository->horarioIncluiDiaCorrente($this->hcmParameters->getNrorgTrabalho(), $nrHordiam)) {
            $toleranciaDiaSeguinte = DateUtil::adicionaIntervalo(
                $toleranciaDiaSeguinte,
                1,
                DateUtil::DIA
            );
        }
        
        return $dtHrMarc >= $toleranciaDiaSeguinte;
    }

    private function insereGpeMarcacaoponto(int $nrvinculom, \DateTime $marcacao, string $nsr, string $idDia, int $nrMarcacaoPonto) : void {
        $dtAtual = DateUtil::getDataAtual();

        $obj = new $this->gpeMarcacaoponto;

        $obj->setNrvinculom($nrvinculom);
        $obj->setNrmarcacaoponto($nrMarcacaoPonto);
        $obj->setNrorg($this->nrOrg);
        $obj->setNrorginclusao($this->nrOrg);
        $obj->setNrorgultatu($this->nrOrg);
        $obj->setDthrmarcacaoponto($marcacao);
        $obj->setDthrajustada($marcacao);
        $obj->setIdnsrrep($nsr);
        $obj->setNrcoletor(self::IMPORTACAOREP_DEFAULT_NRCOLETOR_DEFAULT);
        $obj->setIddiaocorrencia($idDia);
        $obj->setIdenquadrahorario('N');
        $obj->setIdmarcdesprezada('N');
        $obj->setIdativo('S');
        $obj->setDtinclusao($dtAtual);
        $obj->setDtultatu($dtAtual);

        $this->entityManager->persist($obj);
        $this->entityManager->flush($obj);
    }

    private function trataHorarioMarcacoes(array $horarioMarcacoes, int $nrvinculom, int $nrMarcacaoPonto, array $valor) : void {
        if (isset($horarioMarcacoes)) {
            $intervalosDescanso = [];

            // Percorre as marcações inseridas para detectar intervalos
            foreach ($horarioMarcacoes as $marcacoesDia) {
                $quantidadeMarcacoes = count($marcacoesDia);
                $maiorTempoIntervaloDescanso = [];
                
                // Busca índices que estão relacionados com intervalos, considerando um array ordenado de marcações.
                // Ex: 1° Marcação = Entrada; 2° Marcação = Saída; 3° Marcação = Entrada; 4° Marcação = Saída. O índice de um intervalo seria 1 (2° Marcação, primeira saída).
                // Ver mais detalhes na descrição da função

                $possiveisIntervalosDescanso = $this->getIndexesPossiveisDescanso($quantidadeMarcacoes);
                foreach ($possiveisIntervalosDescanso as $index) {
                    $currentIntervalo = DateUtil::getIntervaloEmSegundos($marcacoesDia[$index]['dthrmarcacaoponto'], $marcacoesDia[$index+1]['dthrmarcacaoponto']);
                    
                    // Atualiza o maior intervalo
                    if ($currentIntervalo > ($maiorTempoIntervaloDescanso['tempo'] ?? 0)) {
                        $maiorTempoIntervaloDescanso = [
                            'tempo'             => $currentIntervalo,
                            'inicio'            => $marcacoesDia[$index]['dthrmarcacaoponto'],
                            'fim'               => $marcacoesDia[$index+1]['dthrmarcacaoponto'],
                            'nrmarcacaoponto'   => $marcacoesDia[$index]['nrmarcacaoponto'],
                            'data'              => $marcacoesDia[0]['dthrmarcacaoponto'],
                        ];
                    }
                }
                
                // Guarda o maior intervalo para que a ocorrência seja inserida posteriormente
                if (!empty($maiorTempoIntervaloDescanso)) {
                    $intervalosDescanso[] = $maiorTempoIntervaloDescanso;
                }
            }

            $alteEscala = $this->importacaoAFDRepository->retornaUltimaAlteEscalaPorDataInicio($this->hcmParameters->getNrorgTrabalho(), $nrvinculom, DateUtil::getPrimeiroDiaDoMes($valor['mar'])->format('d/m/Y'));
            
            if(!empty($alteEscala)) {
                $alteEscala  = $alteEscala[0];
                $horaDia     = $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(array("nrorg" => $this->nrOrg, "nrhordiam" => $alteEscala['NRHORDIAM']));
                $nrhordiam   = $alteEscala['NRHORDIAM'];
                $nrintervalo = null;

                foreach($horaDia as $horario) {
                    if ($horario->getHrprojetarintervalo() !== null) {
                        $nrintervalo = $horario->getNrintervalohorario();
                        break;
                    }
                }
                
                // array para definir o id do dia da semana para a tabela
                $diaSemana = ['1' => '0100000',
                              '2' => '0010000',
                              '3' => '0001000',
                              '4' => '0000100',
                              '5' => '0000010',
                              '6' => '0000001',
                              '7' => '1000000'];
                
                // percorre todos os intervalos de almoço salvos e salva uma ocorrência para cada um deles
                foreach ($intervalosDescanso as $intervalo) {
                    
                    $nrmarcacaoponto        = $intervalo['nrmarcacaoponto'];
                    $dataInicioFimProgOcor  = $intervalo['data'];
                    $diaMarcacao = DateUtil::getDiaSemana($dataInicioFimProgOcor->format('d/m/Y'));

                    foreach($diaSemana as $key => $dia) {
                        if($key == $diaMarcacao){
                            $diaMarcacao = $dia;
                        }
                    }

                    $this->inserePtoProgOcorrencia($nrvinculom, $dataInicioFimProgOcor, $diaMarcacao, $nrhordiam, $nrintervalo, $nrMarcacaoPonto);
                }
            }
        }
    }

    /**
     *  Rotina criada com a função de retornar índices de array considerados como inícios de intervalos de descanso a partir da quantidade de marcações
     * em um determinado dia passado como parâmetro
     * 
     *  @param  int     $quantidadeMarcacoes
     * 
     *  @return Array   $nums
     *  Retorna numeros pares > 0 antes do número passado -1. Os números são retornados como índices de array, então como número -1.
     *  Ex. 10: [2, 4, 6] (pares antes de 9) => [1, 3, 5]
     */
    private function getIndexesPossiveisDescanso($quantidadeMarcacoes) {
        $nums=[];
        for ($k=2; $k<$quantidadeMarcacoes-1; $k+=2) {
            $nums[]=$k-1;
        }
        return $nums;
    }

    private function inserePtoProgOcorrencia(int $nrvinculom, \DateTime $dataInicioFimProgOcor, string $diaMarcacao, ?string $nrhordiam, ?int $nrintervalo, int $nrMarcacaoPonto) : void {
        $ocorrencia = new $this->ptoProgOcorrencia;

        $ocorrencia->setNrorg($this->nrOrg);
        $ocorrencia->setIdprogocorconfirmada('S');
        $ocorrencia->setNrvinculom($nrvinculom);
        $ocorrencia->setDtinicioprogocor($dataInicioFimProgOcor);
        $ocorrencia->setDtfimprogocor($dataInicioFimProgOcor);
        $ocorrencia->setIddiassemana($diaMarcacao);
        $ocorrencia->setNrhordiam($nrhordiam);
        $ocorrencia->setNrintervalohorario($nrintervalo);
        $ocorrencia->setIdtipohorario('N');
        $ocorrencia->setNrtipoprogocorrencia(20);
        $ocorrencia->setNrprogocorrencia(\HCM\Util\NovoCodigo::geraCodigo($this->hcmParameters->getNrorgTrabalho(), 'PTO_PROGOCORRENCIA', 1, 12, true));
        $ocorrencia->setNrmarcacaoponto($nrMarcacaoPonto);
        $ocorrencia->setDtinclusao(DateUtil::getDataAtual());
        $ocorrencia->setNrorginclusao($this->hcmParameters->getNrorgTrabalho());
        $ocorrencia->setDtultatu(DateUtil::getDataAtual());
        $ocorrencia->setNrorgultatu($this->hcmParameters->getNrorgTrabalho());

        $this->entityManager->persist($ocorrencia);
        $this->entityManager->flush($ocorrencia);
    }

    /********** FUNÇÕES PARA CONTROLE DIRETO DO REP - CONTROL ID **********/

    public function loginREP(Object $row) : array {
        try {
            $arrResp = [];
            $arrPontoAcesso = $row['NRPONTOACESSO'];
            $dtinifiltro = DateUtil::getDataDeString($row['DTINIFILTRO'], DateUtil::FORMATO_BRASILEIRO_SEM_ZEROS, true);

            foreach($arrPontoAcesso as $nrpontoacesso => $data) {
                $loginResponse = $getSession = $session = $ip = $consulta = $dataAFD = null;
                $ip = $data['IPHOSTREP'];

                $loginResponse = $this->realizaLoginREP($ip)[0];

                if($loginResponse) {
                    //Coletando a session
                    $getSession = json_decode($loginResponse, true);
                    $session = $getSession['session'];

                    $consulta = $this->consultaAFDbyDate($ip, $session, $dtinifiltro)[0];
                    $dataAFD  = base64_decode($consulta);

                    $arrResp[$nrpontoacesso] = ['dspontoacesso' => $data['DSPONTOACESSO'],
                                                'result' => $this->importacaoAFD($dataAFD)];
                }
            }
            return $arrResp;

        } catch (\Exception $e) {
            throw new \Exception("Falha no loginREP AFD. Erro: ". $e->getMessage());
        }
    }

    //Função para realizar chamada HTTPS com o metodo POST 
    private function sendJson(string $url, array $data) : array {
        $ch = curl_init($url);
        $error = null;
        
        $options = [
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_HTTPHEADER => ['Content-Type: application/json'],
            CURLOPT_POST => true,
            CURLOPT_POSTFIELDS => json_encode($data),
            CURLOPT_SSL_VERIFYPEER => false, // Ignorar verificação do certificado SSL
            CURLOPT_SSL_VERIFYHOST => false, // Ignorar verificação do host SSL
        ];
        
        curl_setopt_array($ch, $options);
        
        $response = curl_exec($ch);
        
        if ($response === false) {
            $error = 'cURL error: ' . curl_error($ch);
        }
        
        curl_close($ch);
        
        return [$response] ?? [$error];
    }

    private function realizaLoginREP(string $ip) : array {
        $loginUrl = self::INIT_URL_REP.$ip.self::URL_LOGIN_REP; 

        $loginBody = ['login'    => self::USER_ADM_REP,
                      'password' => self::PASSWORD_ADM_REP];

        $loginResponse = $this->sendJson($loginUrl, $loginBody)[0];	

        if ($loginResponse != false)
            return [$loginResponse];
        else
            return ['Ocorreu um erro ao realizar o login no REP.'];
    }

    private function consultaAFDbyDate(string $ip, string $session, \DateTime $dtinifiltro) : array {
        $loadObjsUrl = self::INIT_URL_REP.$ip.self::URL_GET_AFD_REP;

        $loadObjsBody = ['session' => $session,
                         'initial_date' => [
                            'day'   => $dtinifiltro->format('j'), 
                            'month' => $dtinifiltro->format('n'),
                            'year'  => $dtinifiltro->format('Y')
                        ]];

        $loadObjsResponse = $this->sendJson($loadObjsUrl, $loadObjsBody)[0];	

        if ($loadObjsResponse != false) 
            return [$loadObjsResponse];
        else
            return ['Ocorreu um erro ao consultar o AFD no REP.'];
    }

}